# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from abc import abstractmethod

import jax

import awkward as ak
from awkward import _reducers
from awkward._nplikes.array_like import maybe_materialize
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._nplikes.shape import ShapeItem
from awkward._reducers import Reducer
from awkward._typing import Final, Self, TypeVar

np = NumpyMetadata.instance()


_overloads: dict[type[Reducer], type[JAXReducer]] = {}


R = TypeVar("R", bound=Reducer)


def overloads(cls: type[Reducer]):
    def registrar(new_cls: type[R]) -> type[R]:
        _overloads[cls] = new_cls
        return new_cls

    return registrar


class JAXReducer(Reducer):
    @classmethod
    @abstractmethod
    def from_kernel_reducer(cls, reducer: Reducer) -> Self:
        raise NotImplementedError


def awkward_JAXNumpyArray_reduce_adjust_starts_64(toptr, outlength, parents, starts):
    if outlength == 0 or parents.size == 0:
        return toptr

    identity = jax.numpy.iinfo(jax.numpy.int64).max
    valid = toptr[:outlength] != identity
    safe_sub_toptr = jax.numpy.where(valid, toptr[:outlength], 0)
    parent_indices = parents[safe_sub_toptr]
    adjustments = starts[parent_indices]
    updated = jax.numpy.where(valid, toptr[:outlength] - adjustments, toptr[:outlength])

    return toptr.at[:outlength].set(updated)


def awkward_JAXNumpyArray_reduce_adjust_starts_shifts_64(
    toptr, outlength, parents, starts, shifts
):
    if outlength == 0 or parents.size == 0:
        return toptr

    identity = jax.numpy.iinfo(jax.numpy.int64).max
    valid = toptr[:outlength] != identity
    safe_sub_toptr = jax.numpy.where(valid, toptr[:outlength], 0)
    parent_indices = parents[safe_sub_toptr]
    delta = shifts[safe_sub_toptr] - starts[parent_indices]
    updated = jax.numpy.where(valid, toptr[:outlength] + delta, toptr[:outlength])

    return toptr.at[:outlength].set(updated)


def apply_positional_corrections(
    reduced: ak.contents.NumpyArray,
    parents: ak.index.Index,
    starts: ak.index.Index,
    shifts: ak.index.Index | None,
) -> ak._nplikes.ArrayLike:
    if shifts is None:
        assert (
            parents.nplike is reduced.backend.nplike
            and starts.nplike is reduced.backend.nplike
        )
        return awkward_JAXNumpyArray_reduce_adjust_starts_64(
            reduced.data, reduced.length, parents.data, starts.data
        )

    else:
        assert (
            parents.nplike is reduced.backend.nplike
            and starts.nplike is reduced.backend.nplike
            and shifts.nplike is reduced.backend.nplike
        )
        return awkward_JAXNumpyArray_reduce_adjust_starts_shifts_64(
            reduced.data,
            reduced.length,
            parents.data,
            starts.data,
            shifts.data,
        )


def segment_argmin(data, segment_ids, num_segments):
    """
    Applies a segmented argmin-style reduction.

    Parameters:
        data: jax.numpy.ndarray — the values to reduce.
        segment_ids: same shape as data — indicates segment groupings.
        num_segments: int — total number of segments.

    Returns:
        jax.numpy.ndarray — indices of min within each segment.
    """
    indices = jax.numpy.arange(data.shape[0])

    # Find the minimum value in each segment
    min_vals = jax.ops.segment_min(data, segment_ids, num_segments=num_segments)

    # Find where the data equals the minimum value in each segment
    is_min = data == min_vals[segment_ids]

    # Mask the indices where data matches the minimum value
    masked_indices = jax.numpy.where(is_min, indices, data.shape[0])

    # Return the index of the first minimum value in each segment
    return jax.ops.segment_min(masked_indices, segment_ids, num_segments=num_segments)


@overloads(_reducers.ArgMin)
class ArgMin(JAXReducer):
    name: Final = "argmin"
    needs_position: Final = True
    preferred_dtype: Final = np.int64

    @classmethod
    def from_kernel_reducer(cls, reducer: Reducer) -> Self:
        assert isinstance(reducer, _reducers.ArgMin)
        return cls()

    @classmethod
    def _return_dtype(cls, given_dtype):
        return np.int64

    def apply(
        self,
        array: ak.contents.NumpyArray,
        parents: ak.index.Index,
        starts: ak.index.Index,
        shifts: ak.index.Index | None,
        outlength: ShapeItem,
    ) -> ak.contents.NumpyArray:
        assert isinstance(array, ak.contents.NumpyArray)
        result = segment_argmin(*maybe_materialize(array.data, parents.data), outlength)
        result = jax.numpy.asarray(result, dtype=self.preferred_dtype)
        result_array = ak.contents.NumpyArray(result, backend=array.backend)
        corrected_data = apply_positional_corrections(
            result_array, parents, starts, shifts
        )
        return ak.contents.NumpyArray(corrected_data, backend=array.backend)


def segment_argmax(data, segment_ids, num_segments):
    """
    Applies a segmented argmax-style reduction.

    Parameters:
        data: jax.numpy.ndarray — the values to reduce.
        segment_ids: same shape as data — indicates segment groupings.
        num_segments: int — total number of segments.

    Returns:
        jax.numpy.ndarray — indices of max within each segment.
    """
    indices = jax.numpy.arange(data.shape[0])

    # Find the maximum value in each segment
    max_vals = jax.ops.segment_max(data, segment_ids, num_segments=num_segments)

    # Find where the data equals the maximum value in each segment
    is_max = data == max_vals[segment_ids]

    # Mask the indices where data matches the maximum value
    masked_indices = jax.numpy.where(is_max, indices, data.shape[0])

    # Return the index of the first maximum value in each segment
    return jax.ops.segment_min(masked_indices, segment_ids, num_segments=num_segments)


@overloads(_reducers.ArgMax)
class ArgMax(JAXReducer):
    name: Final = "argmax"
    needs_position: Final = True
    preferred_dtype: Final = np.int64

    @classmethod
    def from_kernel_reducer(cls, reducer: Reducer) -> Self:
        assert isinstance(reducer, _reducers.ArgMax)
        return cls()

    @classmethod
    def _return_dtype(cls, given_dtype):
        return np.int64

    def apply(
        self,
        array: ak.contents.NumpyArray,
        parents: ak.index.Index,
        starts: ak.index.Index,
        shifts: ak.index.Index | None,
        outlength: ShapeItem,
    ) -> ak.contents.NumpyArray:
        assert isinstance(array, ak.contents.NumpyArray)
        result = segment_argmax(*maybe_materialize(array.data, parents.data), outlength)
        result = jax.numpy.asarray(result, dtype=self.preferred_dtype)
        result_array = ak.contents.NumpyArray(result, backend=array.backend)
        corrected_data = apply_positional_corrections(
            result_array, parents, starts, shifts
        )
        return ak.contents.NumpyArray(corrected_data, backend=array.backend)


@overloads(_reducers.Count)
class Count(JAXReducer):
    name: Final = "count"
    preferred_dtype: Final = np.int64
    needs_position: Final = False

    @classmethod
    def from_kernel_reducer(cls, reducer: Reducer) -> Self:
        assert isinstance(reducer, _reducers.Count)
        return cls()

    @classmethod
    def _return_dtype(cls, given_dtype):
        return np.int64

    def apply(
        self,
        array: ak.contents.NumpyArray,
        parents: ak.index.Index,
        starts: ak.index.Index,
        shifts: ak.index.Index | None,
        outlength: ShapeItem,
    ) -> ak.contents.NumpyArray:
        assert isinstance(array, ak.contents.NumpyArray)
        result = jax.numpy.ones_like(
            *maybe_materialize(array.data), dtype=self.preferred_dtype
        )
        result = jax.ops.segment_sum(
            result, *maybe_materialize(parents.data), outlength
        )
        result = jax.numpy.asarray(result, dtype=self.preferred_dtype)

        if np.issubdtype(array.dtype, np.complexfloating):
            return ak.contents.NumpyArray(
                result.view(array.dtype), backend=array.backend
            )
        else:
            return ak.contents.NumpyArray(result, backend=array.backend)


def segment_count_nonzero(data, segment_ids, num_segments):
    """
    Counts the number of non-zero elements in `data` per segment.

    Parameters:
        data: jax.numpy.ndarray — input values to count.
        segment_ids: jax.numpy.ndarray — same shape as data, segment assignment.
        num_segments: int — total number of segments.

    Returns:
        jax.numpy.ndarray — count of non-zero values per segment.
    """
    # Create a binary mask where non-zero entries become 1
    nonzero_mask = jax.numpy.where(data != 0, 1, 0)

    # Sum the mask using segment_sum to count per segment
    return jax.ops.segment_sum(nonzero_mask, segment_ids, num_segments=num_segments)


@overloads(_reducers.CountNonzero)
class CountNonzero(JAXReducer):
    name: Final = "count_nonzero"
    preferred_dtype: Final = np.int64
    needs_position: Final = False

    @classmethod
    def from_kernel_reducer(cls, reducer: Reducer) -> Self:
        assert isinstance(reducer, _reducers.CountNonzero)
        return cls()

    @classmethod
    def _return_dtype(cls, given_dtype):
        return np.int64

    def apply(
        self,
        array: ak.contents.NumpyArray,
        parents: ak.index.Index,
        starts: ak.index.Index,
        shifts: ak.index.Index | None,
        outlength: ShapeItem,
    ) -> ak.contents.NumpyArray:
        assert isinstance(array, ak.contents.NumpyArray)
        result = segment_count_nonzero(
            *maybe_materialize(array.data, parents.data), outlength
        )
        result = jax.numpy.asarray(result, dtype=self.preferred_dtype)

        return ak.contents.NumpyArray(result, backend=array.backend)


@overloads(_reducers.Sum)
class Sum(JAXReducer):
    name: Final = "sum"
    preferred_dtype: Final = np.float64
    needs_position: Final = False

    @classmethod
    def from_kernel_reducer(cls, reducer: Reducer) -> Self:
        assert isinstance(reducer, _reducers.Sum)
        return cls()

    def apply(
        self,
        array: ak.contents.NumpyArray,
        parents: ak.index.Index,
        starts: ak.index.Index,
        shifts: ak.index.Index | None,
        outlength: ShapeItem,
    ) -> ak.contents.NumpyArray:
        assert isinstance(array, ak.contents.NumpyArray)
        if array.dtype.kind == "M":
            raise TypeError(f"cannot compute the sum (ak.sum) of {array.dtype!r}")

        if array.dtype.kind == "b":
            input_array = array.data.astype(np.int64)
        else:
            input_array = array.data
        result = jax.ops.segment_sum(
            *maybe_materialize(input_array, parents.data), outlength
        )

        if array.dtype.kind == "m":
            return ak.contents.NumpyArray(
                array.backend.nplike.asarray(result, dtype=array.dtype)
            )
        elif np.issubdtype(array.dtype, np.complexfloating):
            return ak.contents.NumpyArray(result.view(array.dtype))
        else:
            return ak.contents.NumpyArray(result, backend=array.backend)


def segment_prod_with_negatives(data, segment_ids, num_segments):
    """
    Computes the product of elements in each segment, handling negatives and booleans.
    Parameters:
        data: jax.numpy.ndarray — input values to reduce.
        segment_ids: jax.numpy.ndarray — same shape as data, segment assignment.
        num_segments: int — total number of segments.
    Returns:
        jax.numpy.ndarray — product of values per segment.
    """
    # Handle boolean arrays
    if data.dtype == jax.numpy.bool_:
        return jax.ops.segment_min(
            data.astype(jax.numpy.int32), segment_ids, num_segments
        )

    # For numeric arrays, handle negative values and zeros
    # Track zeros to set product to zero if any segment has zeros
    is_zero = data == 0
    has_zeros = (
        jax.ops.segment_sum(is_zero.astype(jax.numpy.int32), segment_ids, num_segments)
        > 0
    )

    # Track signs to determine final sign of product
    is_negative = data < 0
    neg_count = jax.ops.segment_sum(
        is_negative.astype(jax.numpy.int32), segment_ids, num_segments
    )
    sign_products = 1 - 2 * (
        neg_count % 2
    )  # +1 for even negatives, -1 for odd negatives

    # Calculate product of absolute values in log space
    log_abs = jax.numpy.log(jax.numpy.where(is_zero, 1.0, jax.numpy.abs(data)))
    log_products = jax.ops.segment_sum(
        jax.numpy.where(is_zero, 0.0, log_abs), segment_ids, num_segments
    )
    abs_products = jax.numpy.exp(log_products)

    # Apply zeros and signs
    product = jax.numpy.where(has_zeros, 0.0, sign_products * abs_products)
    # floating point accuracy doesn't let us directly cast to integers
    if np.issubdtype(data.dtype, np.integer):
        result = jax.numpy.round(product).astype(data.dtype)
    else:
        result = product
    return result


@overloads(_reducers.Prod)
class Prod(JAXReducer):
    name: Final = "prod"
    preferred_dtype: Final = np.float64
    needs_position: Final = False

    @classmethod
    def from_kernel_reducer(cls, reducer: Reducer) -> Self:
        assert isinstance(reducer, _reducers.Prod)
        return cls()

    def apply(
        self,
        array: ak.contents.NumpyArray,
        parents: ak.index.Index,
        starts: ak.index.Index,
        shifts: ak.index.Index | None,
        outlength: ShapeItem,
    ) -> ak.contents.NumpyArray:
        assert isinstance(array, ak.contents.NumpyArray)
        # See issue https://github.com/google/jax/issues/9296
        result = segment_prod_with_negatives(
            *maybe_materialize(array.data, parents.data), outlength
        )

        if np.issubdtype(array.dtype, np.complexfloating):
            return ak.contents.NumpyArray(
                result.view(array.dtype), backend=array.backend
            )
        else:
            return ak.contents.NumpyArray(result, backend=array.backend)


@overloads(_reducers.Any)
class Any(JAXReducer):
    name: Final = "any"
    preferred_dtype: Final = np.bool_
    needs_position: Final = False

    @classmethod
    def from_kernel_reducer(cls, reducer: Reducer) -> Self:
        assert isinstance(reducer, _reducers.Any)
        return cls()

    @classmethod
    def _return_dtype(cls, given_dtype):
        return np.bool_

    @staticmethod
    def _max_initial(initial, type):
        if initial is None:
            if type in (
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ):
                return np.iinfo(type).min
            else:
                return -np.inf

        return initial

    def apply(
        self,
        array: ak.contents.NumpyArray,
        parents: ak.index.Index,
        starts: ak.index.Index,
        shifts: ak.index.Index | None,
        outlength: ShapeItem,
    ) -> ak.contents.NumpyArray:
        assert isinstance(array, ak.contents.NumpyArray)
        result = jax.ops.segment_max(
            *maybe_materialize(array.data, parents.data), outlength
        )
        if array.dtype is not np.dtype(bool):
            result = result.at[result == 0].set(self._max_initial(None, array.dtype))
            result = result > self._max_initial(None, array.dtype)
        result = jax.numpy.asarray(result, dtype=bool)

        return ak.contents.NumpyArray(result, backend=array.backend)


@overloads(_reducers.All)
class All(JAXReducer):
    name: Final = "all"
    preferred_dtype: Final = np.bool_
    needs_position: Final = False

    @classmethod
    def from_kernel_reducer(cls, reducer: Reducer) -> Self:
        assert isinstance(reducer, _reducers.All)
        return cls()

    @classmethod
    def _return_dtype(cls, given_dtype):
        return np.bool_

    def apply(
        self,
        array: ak.contents.NumpyArray,
        parents: ak.index.Index,
        starts: ak.index.Index,
        shifts: ak.index.Index | None,
        outlength: ShapeItem,
    ) -> ak.contents.NumpyArray:
        assert isinstance(array, ak.contents.NumpyArray)
        result = jax.ops.segment_min(
            *maybe_materialize(array.data, parents.data), outlength
        )
        result = jax.numpy.asarray(result, dtype=bool)

        return ak.contents.NumpyArray(result, backend=array.backend)


@overloads(_reducers.Min)
class Min(JAXReducer):
    name: Final = "min"
    preferred_dtype: Final = np.float64
    needs_position: Final = False

    def __init__(self, initial):
        self._initial = initial

    @property
    def initial(self):
        return self._initial

    @classmethod
    def from_kernel_reducer(cls, reducer: Reducer) -> Self:
        assert isinstance(reducer, _reducers.Min)
        return cls(reducer.initial)

    @staticmethod
    def _min_initial(initial, type):
        if initial is None:
            if type in (
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ):
                return np.iinfo(type).max
            else:
                return np.inf

        return initial

    def apply(
        self,
        array: ak.contents.NumpyArray,
        parents: ak.index.Index,
        starts: ak.index.Index,
        shifts: ak.index.Index | None,
        outlength: ShapeItem,
    ) -> ak.contents.NumpyArray:
        assert isinstance(array, ak.contents.NumpyArray)

        result = jax.ops.segment_min(
            *maybe_materialize(array.data, parents.data), outlength
        )
        result = jax.numpy.minimum(result, self._min_initial(self.initial, array.dtype))

        if np.issubdtype(array.dtype, np.complexfloating):
            return ak.contents.NumpyArray(
                array.backend.nplike.asarray(
                    result.view(array.dtype), dtype=array.dtype
                ),
                backend=array.backend,
            )
        else:
            return ak.contents.NumpyArray(result, backend=array.backend)


@overloads(_reducers.Max)
class Max(JAXReducer):
    name: Final = "max"
    preferred_dtype: Final = np.float64
    needs_position: Final = False

    def __init__(self, initial):
        self._initial = initial

    @property
    def initial(self):
        return self._initial

    @classmethod
    def from_kernel_reducer(cls, reducer: Reducer) -> Self:
        assert isinstance(reducer, _reducers.Max)
        return cls(reducer.initial)

    @staticmethod
    def _max_initial(initial, type):
        if initial is None:
            if type in (
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ):
                return np.iinfo(type).min
            else:
                return -np.inf

        return initial

    def apply(
        self,
        array: ak.contents.NumpyArray,
        parents: ak.index.Index,
        starts: ak.index.Index,
        shifts: ak.index.Index | None,
        outlength: ShapeItem,
    ) -> ak.contents.NumpyArray:
        assert isinstance(array, ak.contents.NumpyArray)

        result = jax.ops.segment_max(
            *maybe_materialize(array.data, parents.data), outlength
        )
        result = jax.numpy.maximum(result, self._max_initial(self.initial, array.dtype))

        if np.issubdtype(array.dtype, np.complexfloating):
            return ak.contents.NumpyArray(
                array.backend.nplike.asarray(
                    result.view(array.dtype), dtype=array.dtype
                ),
                backend=array.backend,
            )
        else:
            return ak.contents.NumpyArray(result, backend=array.backend)


def get_jax_reducer(reducer: Reducer) -> Reducer:
    return _overloads[type(reducer)].from_kernel_reducer(reducer)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from abc import abstractmethod

from awkward._nplikes.shape import ShapeItem
from awkward._typing import (
    TYPE_CHECKING,
    Any,
    DType,
    EllipsisType,
    Protocol,
    Self,
    SupportsIndex,
    overload,
)

if TYPE_CHECKING:
    from numpy.typing import DTypeLike


class ArrayLike(Protocol):
    @property
    @abstractmethod
    def dtype(self) -> DType: ...

    @property
    @abstractmethod
    def ndim(self) -> int: ...

    @property
    @abstractmethod
    def shape(self) -> tuple[ShapeItem, ...]: ...

    @property
    @abstractmethod
    def strides(self) -> tuple[ShapeItem, ...]: ...

    @property
    @abstractmethod
    def size(self) -> ShapeItem: ...

    @property
    @abstractmethod
    def nbytes(self) -> ShapeItem: ...

    @property
    @abstractmethod
    def T(self) -> Self: ...

    @abstractmethod
    def __getitem__(
        self,
        key: SupportsIndex
        | slice
        | EllipsisType
        | tuple[SupportsIndex | slice | EllipsisType | Self, ...]
        | Self,
    ) -> Self: ...

    @overload
    def __setitem__(
        self,
        key: slice
        | EllipsisType
        | tuple[SupportsIndex | slice | EllipsisType, ...]
        | Self,
        value: int | float | bool | complex,
    ): ...

    @overload
    def __setitem__(
        self,
        key: SupportsIndex
        | slice
        | EllipsisType
        | tuple[SupportsIndex | slice | EllipsisType | Self, ...]
        | Self,
        value: int | float | bool | complex | Self,
    ): ...

    @abstractmethod
    def __setitem__(self, key, value): ...

    @abstractmethod
    def __bool__(self) -> bool: ...

    @abstractmethod
    def __int__(self) -> int: ...

    @abstractmethod
    def __index__(self) -> int: ...

    @abstractmethod
    def __len__(self) -> int: ...

    @abstractmethod
    def view(self, dtype: DTypeLike) -> Self: ...

    # Scalar UFUNCS
    @abstractmethod
    def __add__(self, other: int | complex | float | Self) -> Self: ...

    @abstractmethod
    def __sub__(self, other: int | complex | float | Self) -> Self: ...

    @abstractmethod
    def __mul__(self, other: int | complex | float | Self) -> Self: ...

    @abstractmethod
    def __truediv__(self, other: int | complex | float | Self) -> Self: ...

    @abstractmethod
    def __floordiv__(self, other: int | complex | float | Self) -> Self: ...

    @abstractmethod
    def __gt__(self, other: int | complex | float | Self) -> Self: ...

    @abstractmethod
    def __lt__(self, other: int | complex | float | Self) -> Self: ...

    @abstractmethod
    def __ge__(self, other: int | complex | float | Self) -> Self: ...

    @abstractmethod
    def __le__(self, other: int | complex | float | Self) -> Self: ...

    @abstractmethod
    def __eq__(self, other: int | complex | float | bool | Self) -> Self:  # type: ignore[override]
        ...

    @abstractmethod
    def __and__(self, other: int | bool | Self) -> Self: ...

    @abstractmethod
    def __or__(self, other: int | bool | Self) -> Self: ...

    @abstractmethod
    def __invert__(self) -> Self: ...


class MaterializableArray(ArrayLike):
    @abstractmethod
    def materialize(self) -> ArrayLike: ...


def maybe_materialize(
    *args: Any,
    type_: type[MaterializableArray]
    | tuple[type[MaterializableArray], ...] = MaterializableArray,
) -> tuple[Any, ...]:
    """
    Returns a tuple where all arguments that are instances of `type_` have been replaced
    by the result of calling their `.materialize()` method.

    Other `ArrayLike` or `Any` arguments are returned unchanged.

    Args:
        *args: Variable length argument list of MaterializableArray or ArrayLike or Any objects.
        type_: The class or tuple of classes to check for materialization. The default is `MaterializableArray`.

    Returns:
        tuple: A tuple where each instance of `type_` is replaced by its materialized form,
        and other ArrayLike or Any objects are returned unchanged.
    """
    return tuple(arg.materialize() if isinstance(arg, type_) else arg for arg in args)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import re

import awkward as ak
from awkward._dispatch import high_level_function

__all__ = ("from_raggedtensor",)


@high_level_function()
def from_raggedtensor(array):
    """
    Args:
        array: (`tensorflow.RaggedTensor`):
            RaggedTensor to convert into an  Awkward Array.

    Converts a TensorFlow RaggedTensor into an Awkward Array.

    If `array` contains any other data types the function raises an error.
    """

    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array)


def _impl(array):
    try:
        # get the flat values
        content = array.flat_values
    except AttributeError as err:
        raise TypeError(
            """only RaggedTensor can be converted to awkward array"""
        ) from err

    # handle gpu and cpu instances separately
    device = content.backing_device

    content = _tensor_to_np_or_cp(content, device)

    # convert flat_values to ak.contents right away
    content = ak.contents.NumpyArray(content)

    # get the offsets
    offsets_arr = []
    for splits in array.nested_row_splits:
        # handle gpu and cpu instances separately
        split = _tensor_to_np_or_cp(splits, device)
        # convert to ak.index
        offset = ak.index.Index64(split)
        offsets_arr.append(offset)

    # if a tensor has one *ragged dimension*
    if len(offsets_arr) == 1:
        result = ak.contents.ListOffsetArray(offsets_arr[0], content)
        return ak.Array(result)

    # if a tensor has multiple *ragged dimensions*
    return ak.Array(_recursive_call(content, offsets_arr, 0))


def _tensor_to_np_or_cp(array, device):
    matched_device = re.match(".*:(CPU|GPU):[0-9]+", device)

    if matched_device is None:
        raise NotImplementedError(
            f"TensorFlow device has an unexpected format: {device!r}"
        )
    elif matched_device.groups()[0] == "GPU":
        try:
            import tensorflow as tf
        except ImportError as err:
            raise ImportError(
                """to use ak.from_raggedtensor, you must install the 'tensorflow' package with:

            pip install tensorflow
    or
            conda install tensorflow"""
            ) from err

        from awkward._nplikes.cupy import Cupy

        cp = Cupy.instance()
        return cp.from_dlpack(tf.experimental.dlpack.to_dlpack(array))
    elif matched_device.groups()[0] == "CPU":
        return array.numpy()


def _recursive_call(content, offsets_arr, count):
    if count == len(offsets_arr) - 2:
        return ak.contents.ListOffsetArray(
            offsets_arr[count],
            ak.contents.ListOffsetArray(offsets_arr[count + 1], content),
        )
    else:
        return ak.contents.ListOffsetArray(
            offsets_arr[count], _recursive_call(content, offsets_arr, count)
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from functools import reduce
from itertools import permutations

import awkward as ak
from awkward._backends.dispatch import backend_of_obj
from awkward._dispatch import high_level_function
from awkward._do import mergeable
from awkward._layout import HighLevelContext, ensure_same_backend, maybe_posaxis
from awkward._namedaxis import (
    NAMED_AXIS_KEY,
    NamedAxesWithDims,
    _get_named_axis,
    _named_axis_to_positional_axis,
    _unify_named_axis,
)
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._nplikes.shape import unknown_length
from awkward._parameters import type_parameters_equal
from awkward._regularize import regularize_axis
from awkward._typing import Sequence
from awkward.contents import Content
from awkward.operations.ak_fill_none import fill_none
from awkward.types.numpytype import primitive_to_dtype

__all__ = ("concatenate",)

np = NumpyMetadata.instance()


@ak._connect.numpy.implements("concatenate")
@high_level_function()
def concatenate(
    arrays, axis=0, *, mergebool=True, highlevel=True, behavior=None, attrs=None
):
    """
    Args:
        arrays: Array-like data (anything #ak.to_layout recognizes).
        axis (int or str): The dimension at which this operation is applied. The
            outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc.
            If a str, it is interpreted as the name of the axis which maps
            to an int if named axes are present. Named axes are attached
            to an array using #ak.with_named_axis and removed with
            #ak.without_named_axis; also see the
            [Named axes user guide](../../user-guide/how-to-array-properties-named-axis.html).
        mergebool (bool): If True, boolean and numeric data can be combined
            into the same buffer, losing information about False vs `0` and
            True vs `1`; otherwise, they are kept in separate buffers with
            distinct types (using an #ak.contents.UnionArray).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Returns an array with `arrays` concatenated. For `axis=0`, this means that
    one whole array follows another. For `axis=1`, it means that the `arrays`
    must have the same lengths and nested lists are each concatenated,
    element for element, and similarly for deeper levels.
    """
    # Dispatch
    if (
        # Is an array with a known backend
        backend_of_obj(arrays, default=None) is not None
    ):
        yield (arrays,)
    else:
        yield arrays

    # Implementation
    return _impl(arrays, axis, mergebool, highlevel, behavior, attrs)


def _merge_as_union(
    contents: Sequence[Content], parameters=None
) -> ak.contents.UnionArray:
    length = sum(c.length for c in contents)
    first = contents[0]
    tags = ak.index.Index8.empty(length, first.backend.nplike)
    index = ak.index.Index64.empty(length, first.backend.nplike)

    offset = 0
    for i, content in enumerate(contents):
        content.backend.maybe_kernel_error(
            content.backend["awkward_UnionArray_filltags_const", tags.dtype.type](
                tags.data, offset, content.length, i
            )
        )
        content.backend.maybe_kernel_error(
            content.backend["awkward_UnionArray_fillindex_count", index.dtype.type](
                index.data, offset, content.length
            )
        )
        offset += content.length

    return ak.contents.UnionArray.simplified(
        tags, index, contents, parameters=parameters
    )


def _impl(arrays, axis, mergebool, highlevel, behavior, attrs):
    # Simple single-array, axis=0 fast-path
    if (
        # Is an array with a known backend
        backend_of_obj(arrays, default=None) is not None
    ):
        # Convert the array to a layout object
        content = ak.operations.to_layout(
            arrays, allow_record=False, primitive_policy="error"
        )
        # Only handle concatenation along `axis=0`
        # Let ambiguous depth arrays fall through
        if maybe_posaxis(content, axis, 1) == 0:
            return ak.operations.ak_flatten._impl(arrays, 1, highlevel, behavior, attrs)

    # Now that we're sure `arrays` is not a singular array
    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        content_or_others = ensure_same_backend(
            *(
                ctx.unwrap(
                    x,
                    allow_record=axis != 0,
                    allow_unknown=False,
                    primitive_policy="pass-through",
                )
                for x in arrays
            )
        )

    # Handle named axis
    merged_named_axis = reduce(_unify_named_axis, map(_get_named_axis, arrays))
    # Step 1: normalize named axis to positional axis
    axis = _named_axis_to_positional_axis(merged_named_axis, axis)
    axis = regularize_axis(axis, none_allowed=False)
    # Step 2: propagate named axis from input to output,
    #   use strategy "unify" (see: awkward._namedaxis)
    out_named_axis = merged_named_axis

    contents = [x for x in content_or_others if isinstance(x, ak.contents.Content)]
    if len(contents) == 0:
        raise ValueError("need at least one array to concatenate")

    posaxis = maybe_posaxis(contents[0], axis, 1)
    maxdepth = max(
        x.minmax_depth[1]
        for x in content_or_others
        if isinstance(x, ak.contents.Content)
    )
    if posaxis is None or not 0 <= posaxis < maxdepth:
        raise ValueError(
            f"axis={axis} is beyond the depth of this array or the depth of this array "
            "is ambiguous"
        )
    for x in content_or_others:
        if isinstance(x, ak.contents.Content):
            if maybe_posaxis(x, axis, 1) != posaxis:
                raise ValueError(
                    "arrays to concatenate do not have the same depth for negative "
                    f"axis={axis}"
                )

    if posaxis == 0:
        content_or_others = [
            x if isinstance(x, ak.contents.Content) else ak.operations.to_layout([x])
            for x in content_or_others
        ]
        batches = [[content_or_others[0]]]
        for x in content_or_others[1:]:
            batch = batches[-1]
            if ak._do.mergeable(batch[-1], x, mergebool=mergebool):
                batch.append(x)
            else:
                batches.append([x])

        contents = [ak._do.mergemany(b) for b in batches]
        if len(contents) > 1:
            out = _merge_as_union(contents)
        else:
            out = contents[0]

        if isinstance(out, ak.contents.UnionArray):
            out = type(out).simplified(
                out._tags,
                out._index,
                out._contents,
                parameters=out._parameters,
                mergebool=mergebool,
            )

    else:

        def action(inputs, depth, backend, **kwargs):
            if any(
                x.minmax_depth == (1, 1)
                for x in inputs
                if isinstance(x, ak.contents.Content)
            ):
                raise ValueError(
                    "at least one array is not deep enough to concatenate at "
                    f"axis={axis}"
                )

            if depth != posaxis:
                return

            if any(isinstance(x, ak.contents.Content) and x.is_option for x in inputs):
                nextinputs = []
                for x in inputs:
                    if x.is_option and x.content.is_list:
                        empty = ak.to_backend([], backend)
                        nextinputs.append(fill_none(x, empty, axis=0, highlevel=False))
                    else:
                        nextinputs.append(x)
                inputs = nextinputs

            # Ensure the lengths agree, taking known lengths over unknown lengths
            length = None
            for x in inputs:
                if isinstance(x, ak.contents.Content):
                    if length is None:
                        length = x.length
                    elif x.length is unknown_length:
                        continue
                    elif length is unknown_length:
                        length = x.length
                    elif length != x.length:
                        raise ValueError(
                            f"all arrays must have the same length for axis={axis}"
                        )
            assert length is not None

            if all(
                (isinstance(x, ak.contents.Content) and x.is_regular)
                or (isinstance(x, ak.contents.NumpyArray) and x.data.ndim > 1)
                or not isinstance(x, ak.contents.Content)
                for x in inputs
            ):
                regulararrays = []
                sizes = []
                for x in inputs:
                    if isinstance(x, ak.contents.RegularArray):
                        regulararrays.append(x)
                    elif isinstance(x, ak.contents.NumpyArray):
                        regulararrays.append(x.to_RegularArray())
                    else:
                        regulararrays.append(
                            ak.contents.RegularArray(
                                ak.contents.NumpyArray(
                                    backend.nplike.broadcast_to(
                                        backend.nplike.asarray([x]), (length,)
                                    )
                                ),
                                1,
                            )
                        )
                    sizes.append(regulararrays[-1].size)

                if len(regulararrays) < 2**7:
                    prototype = backend.nplike.empty(sum(sizes), dtype=np.int8)
                    tags_cls = ak.index.Index8
                else:
                    prototype = backend.nplike.empty(sum(sizes), dtype=np.int64)
                    tags_cls = ak.index.Index64

                start = 0
                for tag, size in enumerate(sizes):
                    prototype[start : start + size] = tag
                    start += size

                tags = tags_cls(
                    backend.nplike.reshape(
                        backend.nplike.broadcast_to(
                            prototype, (length, prototype.size)
                        ),
                        (-1,),
                    )
                )
                index = ak.contents.UnionArray.regular_index(tags, backend=backend)
                inner = ak.contents.UnionArray.simplified(
                    tags,
                    index,
                    [x._content for x in regulararrays],
                    mergebool=mergebool,
                )

                return (ak.contents.RegularArray(inner, prototype.size),)

            elif all(
                (isinstance(x, ak.contents.Content) and x.is_list)  # Case 1
                or (isinstance(x, ak.contents.NumpyArray) and x.data.ndim > 1)  # Case 2
                or not isinstance(x, ak.contents.Content)  # Case 3: scalar value
                for x in inputs
            ):
                nextinputs = []
                for x in inputs:
                    if isinstance(x, ak.contents.Content):
                        nextinputs.append(x)
                    else:
                        # Treat non-content inputs as scalars.
                        # These become arrays of matching length.
                        nextinputs.append(
                            ak.contents.ListOffsetArray(
                                ak.index.Index64(
                                    backend.nplike.arange(
                                        backend.nplike.shape_item_as_index(length + 1),
                                        dtype=np.int64,
                                    ),
                                    nplike=backend.nplike,
                                ),
                                ak.contents.NumpyArray(
                                    backend.nplike.broadcast_to(
                                        backend.nplike.asarray([x]), (length,)
                                    )
                                ),
                            )
                        )

                counts = backend.nplike.zeros(nextinputs[0].length, dtype=np.int64)
                all_counts = []
                all_flatten = []

                for x in nextinputs:
                    o, f = x._offsets_and_flattened(axis=1, depth=1)
                    c = o.data[1:] - o.data[:-1]
                    backend.nplike.add(counts, c, maybe_out=counts)
                    all_counts.append(c)
                    all_flatten.append(f)

                offsets = backend.nplike.empty(nextinputs[0].length + 1, dtype=np.int64)
                offsets[0] = 0
                backend.nplike.cumsum(counts, maybe_out=offsets[1:])

                offsets = ak.index.Index64(offsets, nplike=backend.nplike)

                if len(nextinputs) < 2**7:
                    tags_cls = ak.index.Index8
                else:
                    tags_cls = ak.index.Index64
                tags, index = ak.contents.UnionArray.nested_tags_index(
                    offsets,
                    [ak.index.Index64(x) for x in all_counts],
                    backend=backend,
                    tags_cls=tags_cls,
                )

                inner = ak.contents.UnionArray.simplified(
                    tags, index, all_flatten, mergebool=mergebool
                )

                return (ak.contents.ListOffsetArray(offsets, inner),)

            else:
                return None

        depth_context, lateral_context = NamedAxesWithDims.prepare_contexts(
            list(arrays)
        )
        out = ak._broadcasting.broadcast_and_apply(
            content_or_others,
            action,
            depth_context=depth_context,
            lateral_context=lateral_context,
            allow_records=True,
            right_broadcast=False,
        )[0]
        # Unify named axes
        out_named_axis = reduce(
            _unify_named_axis, lateral_context[NAMED_AXIS_KEY].named_axis
        )

    wrapped_out = ctx.wrap(
        out,
        highlevel=highlevel,
    )

    # propagate named axis to output
    return ak.operations.ak_with_named_axis._impl(
        wrapped_out,
        named_axis=out_named_axis,
        highlevel=highlevel,
        behavior=ctx.behavior,
        attrs=ctx.attrs,
    )


def _form_has_type(form, type_):
    """
    Args:
        form: content object
        type_: low-level type object

    Returns True if the form satisfies the given type; otherwise False.
    """
    if not type_parameters_equal(form._parameters, type_._parameters):
        return False

    if form.is_unknown:
        return isinstance(type_, ak.types.UnknownType)
    elif form.is_option:
        return isinstance(type_, ak.types.OptionType) and _form_has_type(
            form.content, type_.content
        )
    elif form.is_indexed:
        return _form_has_type(form.content, type_)
    elif form.is_regular:
        return (
            isinstance(type_, ak.types.RegularType)
            and (
                form.size is unknown_length
                or type_.size is unknown_length
                or form.size == type_.size
            )
            and _form_has_type(form.content, type_.content)
        )
    elif form.is_list:
        return isinstance(type_, ak.types.ListType) and _form_has_type(
            form.content, type_.content
        )
    elif form.is_numpy:
        for _ in range(form.purelist_depth - 1):
            if not isinstance(type_, ak.types.RegularType):
                return False
            type_ = type_.content
        return (
            isinstance(type_, ak.types.NumpyType) and form.primitive == type_.primitive
        )
    elif form.is_record:
        if (
            not isinstance(type_, ak.types.RecordType)
            or type_.is_tuple != form.is_tuple
        ):
            return False

        if form.is_tuple:
            return all(
                _form_has_type(c, t)
                for c, t in zip(form.contents, type_.contents, strict=True)
            )
        else:
            return (frozenset(form.fields) == frozenset(type_.fields)) and all(
                _form_has_type(form.content(f), type_.content(f)) for f in type_.fields
            )
    elif form.is_union:
        if len(form.contents) != len(type_.contents):
            return False

        for contents in permutations(form.contents):
            if all(
                _form_has_type(form, type_)
                for form, type_ in zip(contents, type_.contents, strict=True)
            ):
                return True
        return False
    else:
        raise TypeError(form)


# This routine should not try to replicate the merge logic,
# but we can make use of assumptions w.r.t to what the merge will do.
# e.g., merging can add new unions, promote to options, change dtypes of NumPy arrays
def enforce_concatenated_form(layout, form):
    # Merge invariant (drop known-ness)
    if not layout.is_unknown and form.is_unknown:
        raise AssertionError(
            "merge result should never be of an unknown type unless the layout is unknown"
        )
    # Unknowns become canonical forms
    elif layout.is_unknown and not form.is_unknown:
        return form.length_zero_array().to_backend(layout.backend)

    ############## Unions #####################################################
    # Merge invariant (drop union)
    elif layout.is_union and not form.is_union:
        raise AssertionError("merge result should be a union if layout is a union")
    # Add a union
    elif not layout.is_union and form.is_union:
        # Merge invariant (unions are i8-i64)
        if not (form.tags == "i8" and form.index == "i64"):
            raise AssertionError(
                "merge result that forms a union should have i8 tags and i64 index"
            )

        # Non-categoricals can be merged into union
        if (
            layout.is_indexed
            and not layout.is_option
            and layout.parameter("__array__") != "categorical"
        ):
            index = layout.index.to64()
            # Take the content and drop the parameters (we're taking parameters from form!)
            layout_to_merge = layout.content
        # Otherwise, we move into the contents
        else:
            index = ak.index.Index64(
                layout.backend.nplike.arange(layout.length, dtype=np.int64)
            )
            layout_to_merge = layout

        type_ = layout_to_merge.form.type

        # First assume this type is exactly represented in the union.
        # This won't hold true if any (and not all) of the contents are an option
        # Or if there were mergeable (but non-equal type) pairs in the original
        # concatenation that formed this union
        union_has_exact_type = False
        contents = []
        for content_form in form.contents:
            if _form_has_type(content_form, type_):
                contents.insert(
                    0, enforce_concatenated_form(layout_to_merge, content_form)
                )
                union_has_exact_type = True
            else:
                contents.append(
                    content_form.length_zero_array().to_backend(layout.backend)
                )

        # Otherwise, find anything we can merge with
        if not union_has_exact_type:
            contents.clear()

            for content_form in form.contents:
                # TODO check forms mergeable
                content_layout = content_form.length_zero_array().to_backend(
                    layout.backend
                )
                if mergeable(content_layout, layout_to_merge):
                    contents.insert(
                        0, enforce_concatenated_form(layout_to_merge, content_form)
                    )
                else:
                    contents.append(
                        content_form.length_zero_array().to_backend(layout.backend)
                    )

        return ak.contents.UnionArray(
            ak.index.Index8(layout.backend.nplike.zeros(layout.length, dtype=np.int8)),
            index,
            contents,
            parameters=form._parameters,
        )
    # Preserve union
    elif layout.is_union and form.is_union:
        # Merge invariant (unions are i8-i64)
        if not (form.tags == "i8" and form.index == "i64"):
            raise AssertionError(
                "merge result that forms a union should have i8 tags and i64 index"
            )
        if len(form.contents) < len(layout.contents):
            raise AssertionError(
                "merge result should only grow or preserve a union's cardinality"
            )
        form_contents = [
            f.length_zero_array().to_backend(layout.backend) for f in form.contents
        ]
        form_indices = range(len(form_contents))
        for form_projection_indices in permutations(form_indices, len(layout.contents)):
            if all(
                mergeable(c, form_contents[i])
                for c, i in zip(layout.contents, form_projection_indices, strict=True)
            ):
                break
        else:
            raise AssertionError(
                "merge result should be mergeable against some permutation of the layout"
            )

        next_contents = [
            enforce_concatenated_form(c, form.contents[i])
            for c, i in zip(layout.contents, form_projection_indices, strict=True)
        ]
        next_contents.extend(
            [
                form_contents[i]
                for i in (set(form_indices) - set(form_projection_indices))
            ]
        )
        return ak.contents.UnionArray(
            ak.index.Index8(layout.backend.nplike.astype(layout.tags.data, np.int8)),
            layout.index.to64(),
            next_contents,
            parameters=form._parameters,
        )

    ############## Options ####################################################
    # Merge invariant (drop option)
    elif layout.is_option and not form.is_option:
        raise AssertionError("merge result should be an option if layout is an option")
    # Add option
    elif not layout.is_option and form.is_option:
        return enforce_concatenated_form(
            ak.contents.UnmaskedArray.simplified(layout), form
        )
    # Preserve option
    elif layout.is_option and form.is_option:
        if isinstance(form, ak.forms.IndexedOptionForm):
            if form.index != "i64":
                raise AssertionError(
                    "IndexedOptionForm should have i64 for merge results"
                )
            return layout.to_IndexedOptionArray64().copy(
                content=enforce_concatenated_form(layout.content, form.content),
                parameters=form._parameters,
            )
        # Non IndexedOptionArray types require all merge candidates to have same form
        elif isinstance(
            form,
            (ak.forms.ByteMaskedForm, ak.forms.BitMaskedForm, ak.forms.UnmaskedForm),
        ):
            return layout.copy(
                content=enforce_concatenated_form(layout.content, form.content),
                parameters=form._parameters,
            )
        else:
            raise AssertionError

    ############## Indexed ####################################################
    # Merge invariant (drop indexed)
    elif layout.is_indexed and not form.is_indexed:
        raise AssertionError("merge result must be indexed if layout is indexed")
    # Add index
    elif not layout.is_indexed and form.is_indexed:
        return ak.contents.IndexedArray(
            ak.index.Index64(layout.backend.nplike.arange(layout.length)),
            enforce_concatenated_form(layout, form.content),
            parameters=form._parameters,
        )
    # Preserve index
    elif layout.is_indexed and form.is_indexed:
        if form.index != "i64":
            raise AssertionError("merge result must be i64")
        return ak.contents.IndexedArray(
            layout.index.to64(),
            content=enforce_concatenated_form(layout.content, form.content),
            parameters=form._parameters,
        )

    ############## NumPy ######################################################
    elif layout.is_numpy and form.is_numpy:
        if layout.inner_shape != form.inner_shape:
            raise AssertionError("layout must have same inner_shape as merge result")

        return ak.values_astype(
            # HACK: drop parameters from type so that character arrays are supported
            layout.copy(parameters=None),
            to=primitive_to_dtype(form.primitive),
            highlevel=False,
        ).copy(parameters=form._parameters)

    ############## Lists ######################################################
    # Merge invariant (regular to numpy)
    elif layout.is_regular and form.is_numpy:
        raise AssertionError("layout cannot be regular for NumpyForm merge result")
    # Merge invariant (ragged to regular)
    elif not (layout.is_regular or layout.is_numpy) and form.is_regular:
        raise AssertionError("merge result should be ragged if any input is ragged")
    elif layout.is_numpy and form.is_list:
        if len(layout.inner_shape) == 0:
            raise AssertionError("layout must be at least 2D if merge result is a list")
        return enforce_concatenated_form(layout.to_RegularArray(), form)
    elif layout.is_regular and form.is_regular:
        # regular → regular requires same size!
        if layout.size != form.size:
            raise AssertionError(
                "RegularForm must have same size as layout for merge result"
            )
        return layout.copy(
            content=enforce_concatenated_form(layout.content, form.content),
            parameters=form._parameters,
        )
    elif layout.is_regular and form.is_list:
        if isinstance(form, (ak.forms.ListOffsetForm, ak.forms.ListForm)):
            return enforce_concatenated_form(layout.to_ListOffsetArray64(False), form)
        else:
            raise AssertionError
    elif layout.is_list and form.is_list:
        if isinstance(form, ak.forms.ListOffsetForm):
            layout = layout.to_ListOffsetArray64(False)
            return layout.copy(
                content=enforce_concatenated_form(layout.content, form.content),
                parameters=form._parameters,
            )
        elif isinstance(form, ak.forms.ListForm):
            if not (form.starts == "i64" and form.stops == "i64"):
                raise TypeError("ListForm should have i64 for merge results")
            return ak.contents.ListArray(
                layout.starts.to64(),
                layout.stops.to64(),
                enforce_concatenated_form(layout.content, form.content),
                parameters=form._parameters,
            )
        else:
            raise AssertionError

    ############## Records ####################################################
    # Merge invariant (mix record-tuple)
    elif layout.is_record and not form.is_record:
        raise AssertionError("merge result should be a record if layout is a record")
    # Merge invariant (mix record-tuple)
    elif not layout.is_record and form.is_record:
        raise AssertionError(
            "layout result should be a record if merge result is a record"
        )
    elif layout.is_record and form.is_record:
        if frozenset(layout.fields) != frozenset(form.fields):
            raise AssertionError("merge result and form must have matching fields")
        elif layout.is_tuple != form.is_tuple:
            raise AssertionError(
                "merge result and form must both be tuple or record-like"
            )
        return ak.contents.RecordArray(
            [
                enforce_concatenated_form(layout.content(f), form.content(f))
                for f in layout.fields
            ],
            layout._fields,
            length=layout.length,
            parameters=form._parameters,
            backend=layout.backend,
        )
    else:
        raise NotImplementedError



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext
from awkward._nplikes.numpy_like import NumpyMetadata

__all__ = ("unzip",)

np = NumpyMetadata.instance()


@high_level_function()
def unzip(
    array,
    *,
    how=tuple,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        how (type): The type of the returned output. This can be `tuple` or `dict`.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    If the `array` contains tuples or records, this operation splits them
    into a Python tuple (or dict) of arrays, one for each field.

    If the `array` does not contain tuples or records, the single `array`
    is placed in a length 1 Python tuple (or dict).

    For example,

        >>> array = ak.Array([{"x": 1.1, "y": [1]},
        ...                   {"x": 2.2, "y": [2, 2]},
        ...                   {"x": 3.3, "y": [3, 3, 3]}])
        >>> x, y = ak.unzip(array)
        >>> x
        <Array [1.1, 2.2, 3.3] type='3 * float64'>
        >>> y
        <Array [[1], [2, 2], [3, 3, 3]] type='3 * var * int64'>

    The `how` argument determines the structure of the output. Using `how=dict`
    returns a dictionary of arrays instead of a tuple, and let's you round-trip
    through `ak.zip`:

        >>> array = ak.Array([{"x": 1.1, "y": [1]},
        ...                   {"x": 2.2, "y": [2, 2]},
        ...                   {"x": 3.3, "y": [3, 3, 3]}])
        >>> x = ak.unzip(array, how=dict)
        >>> x
        {'x': <Array [1.1, 2.2, 3.3] type='3 * float64'>,
         'y': <Array [[1], [2, 2], [3, 3, 3]] type='3 * var * int64'>}
        >>> assert ak.zip(ak.unzip(array, how=dict), depth_limit=1).to_list() == array.to_list()  # True

    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, how, highlevel, behavior, attrs)


def _impl(array, how, highlevel, behavior, attrs):
    if how not in {tuple, dict}:
        raise ValueError(f"`how` must be `tuple` or `dict`, not {how!r}")

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array, allow_record=True, primitive_policy="error")

    fields = ak.operations.fields(layout)

    def check_for_union(layout, **kwargs):
        if isinstance(layout, (ak.contents.RecordArray, ak.Record)):
            return layout  # don't descend into nested records

        elif layout.is_union:
            for content in layout.contents:
                if set(ak.operations.fields(content)) != set(fields):
                    raise ValueError(
                        "union of different sets of fields, cannot ak.unzip"
                    )

    ak._do.recursively_apply(layout, check_for_union, return_array=False)

    if len(fields) == 0:
        items = (ctx.wrap(layout, highlevel=highlevel, allow_other=True),)
        fields = ["0"]
    else:
        items = (
            ctx.wrap(layout[n], highlevel=highlevel, allow_other=True) for n in fields
        )
    return tuple(items) if how is tuple else dict(zip(fields, items, strict=True))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext
from awkward._namedaxis import (
    NAMED_AXIS_KEY,
    AxisMapping,
    AxisTuple,
    _prepare_named_axis_for_attrs,
)
from awkward._nplikes.numpy_like import NumpyMetadata

__all__ = ("with_named_axis",)

np = NumpyMetadata.instance()


@high_level_function()
def with_named_axis(
    array,
    named_axis: AxisTuple | AxisMapping,
    *,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        named_axis: AxisTuple | AxisMapping: Names to give to the array axis; this assigns
            the `"__named_axis__"` attr. If None, any existing name is unset.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Returns an #ak.Array or #ak.Record (or low-level equivalent, if
    `highlevel=False`) with a new name. This function does not change the
    array in-place. If the new name is None, then the array is returned as it is.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, named_axis, highlevel, behavior, attrs)


def _impl(array, named_axis, highlevel, behavior, attrs):
    # Named axis handling
    if not named_axis:  # no-op, e.g. named_axis is None, (), {}
        return array

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array, allow_record=True)

    _named_axis = _prepare_named_axis_for_attrs(
        named_axis=named_axis,
        ndim=layout.minmax_depth[1],
    )
    # now we're good, set the named axis
    return ctx.with_attr(
        key=NAMED_AXIS_KEY,
        value=_named_axis,
    ).wrap(
        layout,
        highlevel=highlevel,
        allow_other=True,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import fsspec

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext

__all__ = ("to_safetensors",)


@high_level_function()
def to_safetensors(
    array,
    destination,
    *,
    storage_options=None,
    # ak.to_buffers kwargs
    container=None,
    buffer_key="{form_key}-{attribute}",
    form_key="node{id}",
    id_start=0,
    backend=None,
    byteorder=ak._util.native_byteorder,
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        destination (path-like): Name of the output file, file path, or
            remote URL passed to [fsspec.core.url_to_fs](https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.core.url_to_fs)
            for remote writing.
        storage_options (None or dict): Any additional options to pass to
            [fsspec.core.url_to_fs](https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.core.url_to_fs)
            to open a remote file for writing.
        container (dict, optional): Optional mapping to receive the generated buffer
            bytes. If None (default), a temporary container is used and discarded
            after writing.
        buffer_key (str, optional): Format string for naming buffers. May include
            `{form_key}` and `{attribute}` placeholders. Defaults to
            `"{form_key}-{attribute}"`.
        form_key (str, optional): Format string for node forms when generating buffer
            keys. Typically includes `"{id}"`. Defaults to `"node{id}"`.
        id_start (int, optional): Starting index for node numbering. Defaults to `0`.
        backend (str | object, optional): Backend used to convert array data into
            buffers. If None, the default backend is used.
        byteorder (str, optional): Byte order for numeric buffers. Defaults to the
            system's native byte order.

    Returns:
        None
            This function writes the safetensors file to `destination`. If
        `container` is provided, it will be populated with the raw buffer bytes.

    Serialize an Awkward Array to the safetensors format and write it to `destination`.

    Ref: https://huggingface.co/docs/safetensors/.

    This function converts the provided Awkward Array (or array-like object) into raw
    buffers via `ak.to_buffers` and stores them in the safetensors format. Buffer names
    are generated from `buffer_key` and `form_key` templates, allowing downstream
    compatibility or layout reuse.
    The resulting safetensors file includes metadata containing the Awkward `form` and
    array `length`, which are required for `ak.from_safetensors` to reconstruct the array.

    Example:

        >>> import awkward as ak
        >>> arr = ak.Array([[1, 2, 3], [], [4]])
        >>> ak.to_safetensors(arr, "out.safetensors")


    See also #ak.from_safetensors.
    """
    # Implementation
    return _impl(
        array,
        destination,
        storage_options,
        container,
        buffer_key,
        form_key,
        id_start,
        backend,
        byteorder,
    )


def _impl(
    array,
    destination,
    storage_options,
    container,
    buffer_key,
    form_key,
    id_start,
    backend,
    byteorder,
):
    try:
        from safetensors.numpy import save
    except ImportError as err:
        raise ImportError(
            """to use ak.to_safetensors, you must install the 'safetensors' package with:

        pip install safetensors
or
        conda install -c huggingface safetensors"""
        ) from err

    fs, destination = fsspec.core.url_to_fs(destination, **(storage_options or {}))

    with HighLevelContext(behavior=None, attrs=None) as ctx:
        layout = ctx.unwrap(array, allow_record=True, primitive_policy="error")

    layout = ak.ak_to_packed._impl(
        layout,
        highlevel=False,  # doesn't matter, but we can avoid extra wrapping/unwrapping
        behavior=ctx.behavior,
        attrs=ctx.attrs,
    )

    form, length, buffers = ak.ak_to_buffers._impl(
        layout,
        container=container,
        buffer_key=buffer_key,
        form_key=form_key,
        id_start=id_start,
        backend=backend,
        byteorder=byteorder,
    )

    metadata = {
        "form": form.to_json(),
        "length": str(length),
    }

    byts = save(buffers, metadata)
    # save
    try:
        with fs.open(destination, "wb") as f:
            f.write(byts)
    except Exception as err:
        raise RuntimeError(
            f"Failed to write safetensors file to '{destination}': {err}"
        ) from err



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from awkward._meta.meta import Meta
from awkward._nplikes.shape import ShapeItem
from awkward._typing import Generic, JSONSerializable, TypeVar

T = TypeVar("T", bound=Meta)


class RegularMeta(Meta, Generic[T]):
    is_list = True
    is_regular = True

    size: ShapeItem
    _content: T

    def purelist_parameters(self, *keys: str) -> JSONSerializable:
        if self._parameters is not None:
            for key in keys:
                if key in self._parameters:
                    return self._parameters[key]

        return self._content.purelist_parameters(*keys)

    @property
    def purelist_isregular(self) -> bool:
        return self._content.purelist_isregular

    @property
    def purelist_depth(self) -> int:
        if self.parameter("__array__") in ("string", "bytestring"):
            return 1
        else:
            return self._content.purelist_depth + 1

    @property
    def is_identity_like(self) -> bool:
        return False

    @property
    def minmax_depth(self) -> tuple[int, int]:
        if self.parameter("__array__") in ("string", "bytestring"):
            return (1, 1)
        else:
            mindepth, maxdepth = self._content.minmax_depth
            return (mindepth + 1, maxdepth + 1)

    @property
    def branch_depth(self) -> tuple[bool, int]:
        if self.parameter("__array__") in ("string", "bytestring"):
            return (False, 1)
        else:
            branch, depth = self._content.branch_depth
            return (branch, depth + 1)

    @property
    def fields(self):
        return self._content.fields

    @property
    def is_tuple(self) -> bool:
        return self._content.is_tuple

    @property
    def dimension_optiontype(self) -> bool:
        return False

    @property
    def content(self) -> T:
        return self._content



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import warnings

import awkward as ak
from awkward._nplikes.array_like import ArrayLike, maybe_materialize
from awkward._nplikes.array_module import ArrayModuleNumpyLike
from awkward._nplikes.dispatch import register_nplike
from awkward._nplikes.numpy_like import UfuncLike
from awkward._nplikes.virtual import VirtualNDArray
from awkward._typing import Final, cast


@register_nplike
class Jax(ArrayModuleNumpyLike):
    is_eager: Final = True
    supports_structured_dtypes: Final = False
    supports_virtual_arrays: Final = True

    def __init__(self):
        warnings.warn(
            "The JAX backend is deprecated and will be removed in a future release of Awkward Array. "
            "Please plan to migrate your code accordingly.",
            DeprecationWarning,
            stacklevel=2,
        )
        jax = ak.jax.import_jax()
        self._module = jax.numpy

    def prepare_ufunc(self, ufunc: UfuncLike) -> UfuncLike:
        from awkward._connect.jax import get_jax_ufunc

        return get_jax_ufunc(ufunc)

    @property
    def ma(self):
        raise ValueError(
            "JAX arrays cannot have missing values until JAX implements "
            "numpy.ma.MaskedArray"
        )

    @property
    def char(self):
        raise ValueError(
            "JAX arrays cannot do string manipulations until JAX implements numpy.char"
        )

    @property
    def ndarray(self):
        return self._module.ndarray

    @classmethod
    def is_own_array_type(cls, type_: type) -> bool:
        """
        Args:
            type_: object to test

        Return `True` if the given object is a jax buffer, otherwise `False`.

        """
        return cls.is_array_type(type_) or cls.is_tracer_type(type_)

    @classmethod
    def is_array_type(cls, type_: type) -> bool:
        """
        Args:
            type_: object to test

        Return `True` if the given object is a jax buffer, otherwise `False`.

        """
        module, _, _suffix = type_.__module__.partition(".")
        return module == "jaxlib"

    def is_currently_tracing(self) -> bool:
        jax = ak.jax.import_jax()
        return isinstance(self._module.array(1) + 1, jax.core.Tracer)

    @classmethod
    def is_tracer_type(cls, type_: type) -> bool:
        """
        Args:
            type_: object to test

        Return `True` if the given object is a jax tracer, otherwise `False`.

        """
        module, _, _suffix = type_.__module__.partition(".")
        return module == "jax"

    def is_c_contiguous(self, x: ArrayLike) -> bool:
        return True

    def ascontiguousarray(self, x: ArrayLike) -> ArrayLike:
        if isinstance(x, VirtualNDArray) and x.is_materialized:
            return x.materialize()
        else:
            return x

    def strides(self, x: ArrayLike) -> tuple[int, ...]:
        out: tuple[int, ...] = (x.dtype.itemsize,)
        for item in cast(tuple[int, ...], x.shape[-1:0:-1]):
            out = (item * out[0], *out)
        return out

    ############################ ufuncs, need to overwrite those because JAX doesn't support `out=` for ufuncs

    def add(
        self, x1: ArrayLike, x2: ArrayLike, maybe_out: ArrayLike | None = None
    ) -> ArrayLike:
        del maybe_out
        x1, x2 = maybe_materialize(x1, x2)
        return self._module.add(x1, x2)

    def logical_or(
        self, x1: ArrayLike, x2: ArrayLike, *, maybe_out: ArrayLike | None = None
    ) -> ArrayLike:
        del maybe_out
        x1, x2 = maybe_materialize(x1, x2)
        return self._module.logical_or(x1, x2)

    def logical_and(
        self, x1: ArrayLike, x2: ArrayLike, *, maybe_out: ArrayLike | None = None
    ) -> ArrayLike:
        del maybe_out
        x1, x2 = maybe_materialize(x1, x2)
        return self._module.logical_and(x1, x2)

    def logical_not(
        self, x: ArrayLike, maybe_out: ArrayLike | None = None
    ) -> ArrayLike:
        del maybe_out
        (x,) = maybe_materialize(x)
        return self._module.logical_not(x)

    def sqrt(self, x: ArrayLike, maybe_out: ArrayLike | None = None) -> ArrayLike:
        del maybe_out
        (x,) = maybe_materialize(x)
        return self._module.sqrt(x)

    def exp(self, x: ArrayLike, maybe_out: ArrayLike | None = None) -> ArrayLike:
        del maybe_out
        (x,) = maybe_materialize(x)
        return self._module.exp(x)

    def divide(
        self, x1: ArrayLike, x2: ArrayLike, maybe_out: ArrayLike | None = None
    ) -> ArrayLike:
        del maybe_out
        x1, x2 = maybe_materialize(x1, x2)
        return self._module.divide(x1, x2)

    def memory_ptr(self, x: ArrayLike) -> int:
        (x,) = maybe_materialize(x)
        return x.unsafe_buffer_pointer()  # type: ignore[attr-defined]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

# We're renaming awkward._prettyprint to this module, and gently deprecating the
# private submodule.
from awkward.prettyprint import (
    Formatter,
    FormatterOptions,
    FormatterType,
    PlaceholderValue,
    alternate,
    custom_str,
    get_at,
    get_field,
    half,
    is_identifier,
    valuestr,
    valuestr_horiz,
)

__all__ = [
    "Formatter",
    "FormatterOptions",
    "FormatterType",
    "PlaceholderValue",
    "alternate",
    "custom_str",
    "get_at",
    "get_field",
    "half",
    "is_identifier",
    "valuestr",
    "valuestr_horiz",
]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from awkward._dispatch import high_level_function
from awkward._nplikes.numpy_like import NumpyMetadata

__all__ = ("from_arrow_schema",)

np = NumpyMetadata.instance()


@high_level_function()
def from_arrow_schema(schema):
    """
    Args:
        schema (`pyarrow.Schema`): Apache Arrow schema to convert into an Awkward Form.

    Converts an Apache Arrow schema into an Awkward Form.

    Because awkward uses numpy's dtype system, timestamp types do not have timezones.
    If encountering timestamp types with timezones in the input arrow data, they
    will be silently dropped.

    See also #ak.to_arrow, #ak.to_arrow_table, #ak.from_arrow, #ak.to_parquet, #ak.from_parquet.
    """
    return _impl(schema)


def _impl(schema):
    import awkward._connect.pyarrow

    return awkward._connect.pyarrow.form_handle_arrow(schema, pass_empty_field=True)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import importlib
import sys
import warnings
from functools import reduce

from packaging.version import parse as parse_version

import awkward as ak
from awkward._attrs import attrs_of_obj
from awkward._behavior import behavior_of, behavior_of_obj
from awkward._layout import wrap_layout
from awkward._namedaxis import NAMED_AXIS_KEY, NamedAxesWithDims, _unify_named_axis

_has_checked_version = False
_numexpr_version_ge_2_11_0 = False


def _import_numexpr():
    global _has_checked_version
    global _numexpr_version_ge_2_11_0
    try:
        import numexpr
    except ModuleNotFoundError as err:
        raise ModuleNotFoundError(
            """install the 'numexpr' package with:

    pip install numexpr --upgrade

or

    conda install numexpr"""
        ) from err
    else:
        if not _has_checked_version:
            _numexpr_version = importlib.metadata.version("numexpr")
            if parse_version(_numexpr_version) < parse_version("2.7.1"):
                warnings.warn(
                    "Awkward Array is only known to work with numexpr 2.7.1 or later"
                    f"(you have version {_numexpr_version})",
                    RuntimeWarning,
                    stacklevel=1,
                )
            _has_checked_version = True
            if parse_version(_numexpr_version) >= parse_version("2.11.0"):
                _numexpr_version_ge_2_11_0 = True
        return numexpr


def getArguments(names, local_dict=None, global_dict=None):
    call_frame = sys._getframe(2)

    clear_local_dict = False
    if local_dict is None:
        local_dict = call_frame.f_locals
        clear_local_dict = True
    try:
        frame_globals = call_frame.f_globals
        if global_dict is None:
            global_dict = frame_globals

        clear_local_dict = clear_local_dict and frame_globals is not local_dict

        arguments = []
        for name in names:
            try:
                a = local_dict[name]
            except KeyError:
                a = global_dict[name]
            arguments.append(a)  # <--- different from NumExpr
    finally:
        if clear_local_dict and hasattr(local_dict, "clear"):
            local_dict.clear()

    return arguments


def evaluate(
    expression, local_dict=None, global_dict=None, order="K", casting="safe", **kwargs
):
    numexpr = _import_numexpr()

    if _numexpr_version_ge_2_11_0:
        if not hasattr(numexpr.necompiler._numexpr_last, "l"):
            numexpr.necompiler._numexpr_last.l = numexpr.utils.ContextDict()
        if not hasattr(numexpr.necompiler._names_cache, "c"):
            numexpr.necompiler._names_cache.c = numexpr.utils.CacheDict(256)
        if not hasattr(numexpr.necompiler._numexpr_cache, "c"):
            numexpr.necompiler._numexpr_cache.c = numexpr.utils.CacheDict(256)
        _numexpr_last = numexpr.necompiler._numexpr_last.l
        _names_cache = numexpr.necompiler._names_cache.c
        _numexpr_cache = numexpr.necompiler._numexpr_cache.c
    else:
        _numexpr_last = numexpr.necompiler._numexpr_last
        _names_cache = numexpr.necompiler._names_cache
        _numexpr_cache = numexpr.necompiler._numexpr_cache

    context = numexpr.necompiler.getContext(kwargs)
    expr_key = (expression, tuple(sorted(context.items())))
    if expr_key not in _names_cache:
        _names_cache[expr_key] = numexpr.necompiler.getExprNames(expression, context)
    names, _ex_uses_vml = _names_cache[expr_key]
    arguments = getArguments(names, local_dict, global_dict)

    arrays = [ak.operations.to_layout(x, allow_unknown=True) for x in arguments]

    def action(inputs, **ignore):
        if all(
            isinstance(x, ak.contents.NumpyArray)
            or not isinstance(x, ak.contents.Content)
            for x in inputs
        ):
            input_primitives = [
                x.data if isinstance(x, ak.contents.NumpyArray) else x for x in inputs
            ]
            return (
                ak.contents.NumpyArray(
                    numexpr.evaluate(
                        expression,
                        dict(zip(names, input_primitives, strict=True)),
                        {},
                        order=order,
                        casting=casting,
                        **kwargs,
                    )
                ),
            )
        else:
            return None

    behavior = behavior_of(*arrays)
    depth_context, lateral_context = NamedAxesWithDims.prepare_contexts(arguments)
    out = ak._broadcasting.broadcast_and_apply(
        arrays,
        action,
        depth_context=depth_context,
        lateral_context=lateral_context,
        allow_records=False,
    )
    assert isinstance(out, tuple) and len(out) == 1
    wrapped = wrap_layout(out[0], behavior)
    out_named_axis = reduce(
        _unify_named_axis, lateral_context[NAMED_AXIS_KEY].named_axis
    )
    return ak.operations.ak_with_named_axis._impl(
        wrapped,
        named_axis=out_named_axis,
        highlevel=True,
        behavior=behavior_of_obj(wrapped),
        attrs=attrs_of_obj(wrapped),
    )


evaluate.evaluate = evaluate


def re_evaluate(local_dict=None):
    numexpr = _import_numexpr()

    if _numexpr_version_ge_2_11_0:
        if not hasattr(numexpr.necompiler._numexpr_last, "l"):
            numexpr.necompiler._numexpr_last.l = numexpr.utils.ContextDict()
        _numexpr_last = numexpr.necompiler._numexpr_last.l
    else:
        _numexpr_last = numexpr.necompiler._numexpr_last

    try:
        compiled_ex = _numexpr_last["ex"]  # noqa: F841
    except KeyError as err:
        raise RuntimeError("not a previous evaluate() execution found") from err
    names = _numexpr_last["argnames"]
    arguments = getArguments(names, local_dict)

    arrays = [ak.operations.to_layout(x, allow_unknown=True) for x in arguments]

    def action(inputs, **ignore):
        if all(
            isinstance(x, ak.contents.NumpyArray)
            or not isinstance(x, ak.contents.Content)
            for x in inputs
        ):
            input_primitives = [
                x.data if isinstance(x, ak.contents.NumpyArray) else x for x in inputs
            ]
            return (
                ak.contents.NumpyArray(
                    numexpr.re_evaluate(dict(zip(names, input_primitives, strict=True)))
                ),
            )
        else:
            return None

    behavior = behavior_of(*arrays)

    depth_context, lateral_context = NamedAxesWithDims.prepare_contexts(arguments)
    out = ak._broadcasting.broadcast_and_apply(
        arrays,
        action,
        depth_context=depth_context,
        lateral_context=lateral_context,
        allow_records=False,
    )
    assert isinstance(out, tuple) and len(out) == 1
    wrapped = wrap_layout(out[0], behavior)
    out_named_axis = reduce(
        _unify_named_axis, lateral_context[NAMED_AXIS_KEY].named_axis
    )
    return ak.operations.ak_with_named_axis._impl(
        wrapped,
        named_axis=out_named_axis,
        highlevel=True,
        behavior=behavior_of_obj(wrapped),
        attrs=attrs_of_obj(wrapped),
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from collections.abc import Mapping

import awkward as ak
from awkward._typing import Any
from awkward._util import STDOUT
from awkward.types.type import Type


class ScalarType:
    def __init__(self, content: Type, behavior: Mapping | None = None):
        if not isinstance(content, ak.types.Type):
            raise TypeError(
                f"{type(self).__name__} all 'contents' must be Type subclasses, not {content!r}"
            )
        self._content: Type = content
        self._behavior: Mapping | None = behavior

    @property
    def content(self) -> Type:
        return self._content

    @property
    def behavior(self) -> Mapping | None:
        return self._behavior

    def __str__(self) -> str:
        return "".join(self._str("", True))

    def show(self, stream=STDOUT):
        out = "".join(self._str("", False))
        if out is None:
            return out
        else:
            if stream is STDOUT:
                stream = STDOUT.stream
            stream.write(out + "\n")

    def _str(self, indent: str, compact: bool) -> list[str]:
        return self._content._str(
            indent,
            compact,
            self._behavior,
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._content!r}, {self._behavior!r})"

    def is_equal_to(self, other: Any, *, all_parameters: bool = False) -> bool:
        return isinstance(other, type(self)) and self._content.is_equal_to(
            other._content, all_parameters=all_parameters
        )

    __eq__ = is_equal_to



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import fsspec

import awkward as ak
from awkward._dispatch import high_level_function

__all__ = ("from_safetensors",)


@high_level_function()
def from_safetensors(
    source,
    *,
    storage_options=None,
    virtual=False,
    # ak.from_buffers kwargs
    buffer_key="{form_key}-{attribute}",
    backend="cpu",
    byteorder="<",
    allow_noncanonical_form=False,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        source (path-like): Name of the input file, file path, or
            remote URL passed to [fsspec.core.url_to_fs](https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.core.url_to_fs)
            for remote reading.
        storage_options (None or dict): Any additional options to pass to
            [fsspec.core.url_to_fs](https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.core.url_to_fs)
            to open a remote file for reading.
        virtual (bool, optional): If True, create a virtual (lazy) Awkward Array
           that references buffers without materializing them. Defaults to False.
        buffer_key (str, optional): Template for buffer names, with placeholders
           `{form_key}` and `{attribute}`. Defaults to "{form_key}-{attribute}".
        backend (str, optional): Backend identifier (e.g., "cpu"). Defaults to "cpu".
        byteorder (str, optional): Byte order, "<" (little-endian, default) or ">".
        allow_noncanonical_form (bool, optional): If True, normalize
            safetensors forms that do not directly match Awkward. Defaults to False.
         highlevel (bool, optional): If True, return a high-level ak.Array. If False,
             return the low-level layout. Defaults to True.
         behavior (Mapping | None, optional): Optional Awkward behavior mapping.
         attrs (Mapping | None, optional): Optional metadata to attach to the array.

    Returns:
        ak.Array or ak.layout.Content: An Awkward Array (or layout) reconstructed
        from the safetensors buffers.

    Load a safetensors file as an Awkward Array.

    Ref: https://huggingface.co/docs/safetensors/.

    This function reads data serialized in the safetensors format and reconstructs
    an Awkward Array (or low-level layout) from it. Buffers in the safetensors file
    are mapped to Awkward buffers according to the `buffer_key` template, and
    optional behavior or attributes can be attached to the returned array.

    The safetensors file **must contain** `form` and `length` entries in its
    metadata, which define the structure and length of the reconstructed array.

    Example:

        >>> import awkward as ak
        >>> arr = ak.from_safetensors("out.safetensors")
        >>> arr  # doctest: +SKIP
        <Array [[1, 2, 3], [], [4]] type='3 * var * int64'>

        Create a virtual (lazy) array that references buffers without materializing them:

        >>> virtual_arr = ak.from_safetensors("out.safetensors", virtual=True)
        >>> virtual_arr  # doctest: +SKIP
        <Array [??, ??, ??] type='3 * var * int64'>


    See also #ak.to_safetensors.
    """
    # Implementation
    return _impl(
        source,
        storage_options,
        virtual,
        buffer_key,
        backend,
        byteorder,
        allow_noncanonical_form,
        highlevel,
        behavior,
        attrs,
    )


def _impl(
    source,
    storage_options,
    virtual,
    buffer_key,
    backend,
    byteorder,
    allow_noncanonical_form,
    highlevel,
    behavior,
    attrs,
):
    try:
        from safetensors import _safe_open_handle
    except ImportError as err:
        raise ImportError(
            """to use ak.from_tensorflow, you must install the 'safetensors' package with:

        pip install safetensors
or
        conda install -c huggingface safetensors"""
        ) from err

    fs, source = fsspec.core.url_to_fs(source, **(storage_options or {}))

    buffers = {}

    def maybe_virtualize(x):
        return (lambda: x) if virtual else x

    with fs.open(source, "rb") as f:
        with _safe_open_handle(f, framework="np") as g:
            metadata = g.metadata()
            for k in g.offset_keys():
                buffers[k] = maybe_virtualize(g.get_tensor(k))

    if "form" not in metadata or "length" not in metadata:
        raise RuntimeError(
            "Missing required metadata in safetensors file: 'form' and 'length' are required."
        )
    form = ak.forms.from_json(metadata["form"])
    length = int(metadata["length"])

    # reconstruct array
    return ak.ak_from_buffers._impl(
        form,
        length,
        buffers,
        buffer_key=buffer_key,
        backend=backend,
        byteorder=byteorder,
        simplify=allow_noncanonical_form,
        enable_virtualarray_caching=True,
        highlevel=highlevel,
        behavior=behavior,
        attrs=attrs,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import itertools
import json
import math
import re
from collections import defaultdict
from collections.abc import Callable, Iterable, Mapping
from fnmatch import fnmatchcase
from glob import escape as escape_glob

import awkward as ak
from awkward._backends.dispatch import regularize_backend
from awkward._backends.numpy import NumpyBackend
from awkward._meta.meta import Meta
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._nplikes.shape import unknown_length
from awkward._parameters import (
    parameters_are_equal,
    parameters_union,
    type_parameters_equal,
)
from awkward._typing import (
    Any,
    DType,
    Final,
    Iterator,
    JSONMapping,
    Self,
    Tuple,
    TypeAlias,
)

__all__ = ("Form", "from_dict", "from_json", "from_type", "reserved_nominal_parameters")

np = NumpyMetadata.instance()
numpy_backend = NumpyBackend.instance()

FormKeyPathT: TypeAlias = Tuple[str | int | None, ...]

reserved_nominal_parameters: Final = frozenset(
    {
        ("__array__", "string"),
        ("__array__", "bytestring"),
        ("__array__", "char"),
        ("__array__", "byte"),
        ("__array__", "sorted_map"),
        ("__array__", "categorical"),
    }
)


def from_dict(input: Mapping) -> Form:
    assert input is not None
    if isinstance(input, str):
        return ak.forms.NumpyForm(primitive=input)

    assert isinstance(input, Mapping)
    parameters = input.get("parameters", None)
    form_key = input.get("form_key", None)

    if input["class"] == "NumpyArray":
        primitive = input["primitive"]
        inner_shape = tuple(
            unknown_length if item is None else item
            for item in input.get("inner_shape", [])
        )
        return ak.forms.NumpyForm(
            primitive, inner_shape, parameters=parameters, form_key=form_key
        )

    elif input["class"] == "EmptyArray":
        return ak.forms.EmptyForm(parameters=parameters, form_key=form_key)

    elif input["class"] == "RegularArray":
        return ak.forms.RegularForm(
            content=from_dict(input["content"]),
            size=unknown_length if input["size"] is None else input["size"],
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] in ("ListArray", "ListArray32", "ListArrayU32", "ListArray64"):
        return ak.forms.ListForm(
            starts=input["starts"],
            stops=input["stops"],
            content=from_dict(input["content"]),
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] in (
        "ListOffsetArray",
        "ListOffsetArray32",
        "ListOffsetArrayU32",
        "ListOffsetArray64",
    ):
        return ak.forms.ListOffsetForm(
            offsets=input["offsets"],
            content=from_dict(input["content"]),
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] == "RecordArray":
        # New serialisation
        if "fields" in input:
            if isinstance(input["contents"], Mapping):
                raise TypeError("new-style RecordForm contents must not be mappings")
            contents = [from_dict(content) for content in input["contents"]]
            fields = input["fields"]
        # Old style record
        elif isinstance(input["contents"], Mapping):
            contents = []
            fields = []
            for key, content in input["contents"].items():
                contents.append(from_dict(content))
                fields.append(key)
        # Old style tuple
        else:
            contents = [from_dict(content) for content in input["contents"]]
            fields = None
        return ak.forms.RecordForm(
            contents=contents,
            fields=fields,
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] in (
        "IndexedArray",
        "IndexedArray32",
        "IndexedArrayU32",
        "IndexedArray64",
    ):
        return ak.forms.IndexedForm(
            index=input["index"],
            content=from_dict(input["content"]),
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] in (
        "IndexedOptionArray",
        "IndexedOptionArray32",
        "IndexedOptionArray64",
    ):
        return ak.forms.IndexedOptionForm(
            index=input["index"],
            content=from_dict(input["content"]),
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] == "ByteMaskedArray":
        return ak.forms.ByteMaskedForm(
            mask=input["mask"],
            content=from_dict(input["content"]),
            valid_when=input["valid_when"],
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] == "BitMaskedArray":
        return ak.forms.BitMaskedForm(
            mask=input["mask"],
            content=from_dict(input["content"]),
            valid_when=input["valid_when"],
            lsb_order=input["lsb_order"],
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] == "UnmaskedArray":
        return ak.forms.UnmaskedForm(
            content=from_dict(input["content"]),
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] in (
        "UnionArray",
        "UnionArray8_32",
        "UnionArray8_U32",
        "UnionArray8_64",
    ):
        return ak.forms.UnionForm(
            tags=input["tags"],
            index=input["index"],
            contents=[from_dict(content) for content in input["contents"]],
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] == "VirtualNDArray":
        raise ValueError("Awkward 1.x VirtualNDArrays are not supported")

    else:
        raise ValueError(
            "input class: {} was not recognised".format(repr(input["class"]))
        )


def from_json(input: str) -> Form:
    return from_dict(json.loads(input))


def from_type(type_: ak.types.Type) -> Form:
    # Categorical types are reintroduced into forms using metadata
    if type_.parameter("__categorical__"):
        # Drop categorical placeholder parameter
        if type_._parameters is None:
            next_parameters = None
        else:
            next_parameters = type_._parameters.copy()
            next_parameters.pop("__categorical__")

        if isinstance(type_, ak.types.OptionType):
            next_content = from_type(type_.content)
            return ak.forms.IndexedOptionForm(
                "i64",
                next_content,
                parameters=parameters_union(
                    next_parameters, {"__array__": "categorical"}
                ),
            )
        else:
            next_content = from_type(type_.copy(parameters=next_parameters))
            return ak.forms.IndexedForm(
                "i64", next_content, parameters={"__array__": "categorical"}
            )

    if isinstance(type_, ak.types.NumpyType):
        return ak.forms.NumpyForm(type_.primitive, parameters=type_._parameters)
    elif isinstance(type_, ak.types.ListType):
        return ak.forms.ListOffsetForm(
            "i64", from_type(type_.content), parameters=type_._parameters
        )
    elif isinstance(type_, ak.types.RegularType):
        return ak.forms.RegularForm(
            from_type(type_.content),
            size=type_.size,
            parameters=type_._parameters,
        )
    elif isinstance(type_, ak.types.OptionType):
        return ak.forms.IndexedOptionForm(
            "i64",
            from_type(type_.content),
            parameters=type_._parameters,
        )
    elif isinstance(type_, ak.types.RecordType):
        return ak.forms.RecordForm(
            [from_type(c) for c in type_.contents],
            type_.fields,
            parameters=type_._parameters,
        )
    elif isinstance(type_, ak.types.UnionType):
        return ak.forms.UnionForm(
            "i8",
            "i64",
            [from_type(c) for c in type_.contents],
            parameters=type_._parameters,
        )
    elif isinstance(type_, ak.types.UnknownType):
        return ak.forms.EmptyForm(parameters=type_._parameters)
    elif isinstance(type_, (ak.types.ArrayType, ak.types.ScalarType)):
        raise TypeError(
            "High-level types (ak.types.ArrayType, ak.types.ScalarType) do not have representations as Awkward forms. "
            "Instead the low level type should be used."
        )
    else:
        raise TypeError(f"unsupported type {type_!r}")


def _expand_braces(text, seen=None):
    if seen is None:
        seen = set()

    spans = [m.span() for m in re.finditer(r"\{[^\{\}]*\}", text)][::-1]
    alts = [text[start + 1 : stop - 1].split(",") for start, stop in spans]

    if len(spans) == 0:
        if text not in seen:
            yield text
        seen.add(text)

    else:
        for combo in itertools.product(*alts):
            replaced = list(text)
            for (start, stop), replacement in zip(spans, combo, strict=True):
                replaced[start:stop] = replacement
            yield from _expand_braces("".join(replaced), seen)


class _SpecifierMatcher:
    def __init__(
        self, specifiers: Iterable[list[str]], *, match_if_empty: bool = False
    ):
        # We'll build two sets of unique fixed-strings and patterns
        fixed_strings = set()
        patterns = set()
        # And then map these unique strings to their child specifiers
        match_to_next_specifiers: defaultdict[str, list[list[str]]] = defaultdict(list)

        # For each specifier, categorise it as a fixed-string or pattern,
        # and build the next-specifier table
        parent: str
        child: list[str]
        for item in specifiers:
            parent, *child = item

            if escape_glob(parent) == parent:
                fixed_strings.add(parent)
            else:
                patterns.add(parent)

            # Only include child specifier list if it is non-empty
            if child:
                match_to_next_specifiers[parent].append(child)

        self._match_to_next_specifiers = match_to_next_specifiers
        self._fixed_strings = fixed_strings
        self._patterns = patterns
        self._match_if_empty = match_if_empty

    @property
    def is_empty(self) -> bool:
        return not (self._patterns or self._fixed_strings)

    def __call__(self, field: str, *, next_match_if_empty: bool = False) -> Self | None:
        has_matched = False

        # Fixed-strings are an O(log n) lookup
        next_specifiers = []
        if field in self._fixed_strings:
            has_matched = True
            next_specifiers.extend(self._match_to_next_specifiers[field])

        # Patterns are an O(n) lookup
        for pattern in self._patterns:
            if fnmatchcase(field, pattern):
                has_matched = True
                next_specifiers.extend(self._match_to_next_specifiers[pattern])

        if has_matched:
            return type(self)(next_specifiers, match_if_empty=next_match_if_empty)
        elif self.is_empty and self._match_if_empty:
            return self
        else:
            return None


def regularize_buffer_key(buffer_key: str | Callable) -> Callable[[Form, str], str]:
    if isinstance(buffer_key, str):

        def getkey(form, attribute):
            return buffer_key.format(form_key=form.form_key, attribute=attribute)

        return getkey

    elif callable(buffer_key):

        def getkey(form, attribute):
            return buffer_key(form_key=form.form_key, attribute=attribute, form=form)

        return getkey

    else:
        raise TypeError(
            f"buffer_key must be a string or a callable, not {type(buffer_key)}"
        )


index_to_dtype: Final[dict[str, DType]] = {
    "i8": np.dtype("i1"),
    "u8": np.dtype("u1"),
    "i32": np.dtype("i4"),
    "u32": np.dtype("u4"),
    "i64": np.dtype("i8"),
}


class Form(Meta):
    def _init(self, *, parameters: JSONMapping | None, form_key: str | None):
        if parameters is not None and not isinstance(parameters, dict):
            raise TypeError(
                f"{type(self).__name__} 'parameters' must be of type dict or None, not {parameters!r}"
            )
        if form_key is not None and not isinstance(form_key, str):
            raise TypeError(
                f"{type(self).__name__} 'form_key' must be of type string or None, not {form_key!r}"
            )

        self._parameters = parameters
        self._form_key = form_key

    @property
    def form_key(self):
        return self._form_key

    @form_key.setter
    def form_key(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError("form_key must be None or a string")
        self._form_key = value

    def __str__(self):
        return json.dumps(self.to_dict(verbose=False), indent=4)

    def to_dict(self, verbose=True):
        return self._to_dict_part(verbose, toplevel=True)

    def _to_dict_extra(self, out, verbose):
        if verbose or (self._parameters is not None and len(self._parameters) > 0):
            out["parameters"] = self.parameters
        if verbose or self._form_key is not None:
            out["form_key"] = self._form_key
        return out

    def to_json(self):
        return json.dumps(self.to_dict(verbose=True))

    def _repr_args(self):
        out = []
        if self._parameters is not None and len(self._parameters) > 0:
            out.append("parameters=" + repr(self._parameters))
        if self._form_key is not None:
            out.append("form_key=" + repr(self._form_key))
        return out

    @property
    def type(self):
        raise NotImplementedError

    def columns(self, list_indicator=None, column_prefix=()):
        output = []
        self._columns(column_prefix, output, list_indicator)
        return output

    def select_columns(
        self, specifier, expand_braces=True, *, prune_unions_and_records: bool = True
    ):
        """
        select_columns returns a new Form with only columns and sub-columns selected.
        Returns an empty Form if no columns matched the specifier(s).

        `specifier` can be a `str | Iterable[str | Iterable[str]]`.
        Strings may include shell-globbing-style wildcards "*" and "?".
        If `expand_braces` is `True` (the default), strings may include alternatives in braces.
        For example, `["a.{b,c}.d"]` is equivalent to `["a.b.d", "a.c.d"]`.
        Glob-style matching would also suit this single-character instance: `"a.[bc].d"`.
        If specifier is a list which contains a list/tuple, that inner list will be interpreted as
        column and subcolumn specifiers. They *may* contain wildcards, but "." will not be
        interpreted as a `<field>.<subfield>` pattern.
        """
        if isinstance(specifier, str):
            specifier = {specifier}

        # Only take unique specifiers
        for item in specifier:
            if isinstance(item, str):
                if item == "":
                    raise ValueError(
                        "a column-selection specifier cannot be an empty string"
                    )
            elif isinstance(item, Iterable):
                for field in item:
                    if not isinstance(field, str):
                        raise ValueError("a sub-column specifier must be a string")
            else:
                raise TypeError(
                    "a column specifier must be a string or an iterable of strings"
                )

        if expand_braces:
            next_specifier = []
            for item in specifier:
                if isinstance(item, str):
                    for result in _expand_braces(item):
                        next_specifier.append(result)
                else:
                    next_specifier.append(item)
            specifier = next_specifier

        # specifier = set(specifier)
        specifier_lists: list[list[str]] = []
        for item in specifier:
            if isinstance(item, str):
                if item == "":
                    specifier_lists.append([])
                else:
                    specifier_lists.append(item.split("."))
            else:
                specifier_lists.append(item)
        match_specifier = _SpecifierMatcher(specifier_lists, match_if_empty=False)
        selection = self._select_columns(match_specifier)
        assert selection is not None, "top-level selections always return a Form"

        if prune_unions_and_records:
            return selection._prune_columns(False)
        else:
            return selection

    def column_types(self):
        return self._column_types()

    def _columns(self, path, output, list_indicator):
        raise NotImplementedError

    def _prune_columns(self, is_inside_record_or_union: bool) -> Form | None:
        raise NotImplementedError

    def _select_columns(self, match_specifier: _SpecifierMatcher) -> Form | None:
        raise NotImplementedError

    def _column_types(self):
        raise NotImplementedError

    def _to_dict_part(self, verbose, toplevel):
        raise NotImplementedError

    def length_zero_array(
        self, *, backend=numpy_backend, highlevel=False, behavior=None
    ):
        if highlevel:
            raise ValueError(
                "The `highlevel=True` variant of `Form.length_zero_array` has been removed. "
                "Please use `ak.Array(form.length_zero_array(...), behavior=...)` if an `ak.Array` is required.",
            )
        if not regularize_backend(backend).nplike.known_data:
            return self.length_zero_array(
                backend=numpy_backend, highlevel=highlevel, behavior=behavior
            ).to_typetracer()

        return ak.operations.ak_from_buffers._impl(
            form=self,
            length=0,
            container={"": b"\x00\x00\x00\x00\x00\x00\x00\x00"},
            buffer_key="",
            backend=backend,
            byteorder=ak._util.native_byteorder,
            highlevel=False,
            behavior=behavior,
            attrs=None,
            simplify=False,
            enable_virtualarray_caching=True,
        )

    def length_one_array(
        self, *, backend=numpy_backend, highlevel=False, behavior=None
    ):
        if highlevel:
            raise ValueError(
                "The `highlevel=True` variant of `Form.length_one_array` has been removed. "
                "Please use `ak.Array(form.length_one_array(...), behavior=...)` if an `ak.Array` is required.",
            )
        if not regularize_backend(backend).nplike.known_data:
            return self.length_one_array(
                backend=numpy_backend, highlevel=highlevel, behavior=behavior
            ).to_typetracer()

        container = {}

        def prepare(form, length):
            form_key = f"node-{len(container)}"

            if isinstance(form, ak.forms.EmptyForm):
                if length != 0:
                    raise TypeError(
                        "cannot generate a length_one_array from a Form with an "
                        "unknowntype that cannot be hidden (EmptyForm requires length=0)"
                    )
                return form

            elif isinstance(form, ak.forms.NumpyForm):
                dtype = ak.types.numpytype.primitive_to_dtype(form._primitive)
                size = length * dtype.itemsize
                for x in form.inner_shape:
                    if x is not unknown_length:
                        size *= x
                container[form_key] = b"\x00" * size
                return form.copy(form_key=form_key)

            elif isinstance(form, ak.forms.UnmaskedForm):
                # no buffer, content length = length
                return form.copy(content=prepare(form.content, length))

            elif isinstance(form, ak.forms.BitMaskedForm):
                # mask buffer with values that indicate None
                dtype = index_to_dtype[form.mask]
                mask_length = math.ceil(length / 8.0)
                if form.valid_when:
                    container[form_key] = b"\x00" * (mask_length * dtype.itemsize)
                else:
                    container[form_key] = b"\xff" * (mask_length * dtype.itemsize)
                # content length must equal length
                next_length = length
                return form.copy(
                    content=prepare(form.content, next_length), form_key=form_key
                )

            elif isinstance(form, ak.forms.ByteMaskedForm):
                # mask buffer with values that indicate None
                dtype = index_to_dtype[form.mask]
                if form.valid_when:
                    container[form_key] = b"\x00" * (length * dtype.itemsize)
                else:
                    container[form_key] = b"\x01" * (length * dtype.itemsize)
                # content length must equal mask length
                next_length = length
                return form.copy(
                    content=prepare(form.content, next_length), form_key=form_key
                )

            elif isinstance(form, ak.forms.IndexedOptionForm):
                # index buffer with all -1
                dtype = index_to_dtype[form.index]
                minus_one = b"\xff" * dtype.itemsize
                container[form_key] = minus_one * length
                # with all -1 indices, next_length = max(0, max(index) + 1) = 0
                next_length = 0
                return form.copy(
                    content=prepare(form.content, next_length), form_key=form_key
                )

            elif isinstance(form, ak.forms.IndexedForm):
                # index buffer with all zeros
                dtype = index_to_dtype[form.index]
                container[form_key] = b"\x00" * (length * dtype.itemsize)
                # with all-zero indices, next_length = max(index) + 1 = 1 (if length > 0)
                next_length = 1 if length > 0 else 0
                return form.copy(
                    content=prepare(form.content, next_length), form_key=form_key
                )

            elif isinstance(form, ak.forms.ListForm):
                # starts and stops buffers, both all zeros (share same buffer via form_key)
                # use max itemsize in case starts and stops have different dtypes
                itemsize = max(
                    index_to_dtype[form.starts].itemsize,
                    index_to_dtype[form.stops].itemsize,
                )
                container[form_key] = b"\x00" * (length * itemsize)
                # with starts=stops=0, next_length = max(stops[starts != stops]) = 0
                next_length = 0
                return form.copy(
                    content=prepare(form.content, next_length), form_key=form_key
                )

            elif isinstance(form, ak.forms.ListOffsetForm):
                # offsets buffer has length+1 elements, all zeros
                dtype = index_to_dtype[form.offsets]
                container[form_key] = b"\x00" * ((length + 1) * dtype.itemsize)
                # with all-zero offsets, next_length = offsets[-1] = 0
                next_length = 0
                return form.copy(
                    content=prepare(form.content, next_length), form_key=form_key
                )

            elif isinstance(form, ak.forms.RegularForm):
                size = form.size
                # https://github.com/scikit-hep/awkward/pull/2499#discussion_r1220503454
                if size is unknown_length:
                    size = 1
                next_length = length * size
                return form.copy(content=prepare(form.content, next_length))

            elif isinstance(form, ak.forms.RecordForm):
                # no buffer, all contents have same length
                return form.copy(contents=[prepare(x, length) for x in form.contents])

            elif isinstance(form, ak.forms.UnionForm):
                # tags all 0, index all 0
                dtype_index = index_to_dtype[form.index]
                container[form_key] = b"\x00" * (length * dtype_index.itemsize)
                # content[0]: next_length = max(index[tags==0]) + 1 = 1 (if length > 0)
                # content[i>0]: next_length = 0 (no elements with those tags)
                contents = []
                for i, content_form in enumerate(form.contents):
                    if i == 0:
                        next_length = 1 if length > 0 else 0
                    else:
                        next_length = 0
                    contents.append(prepare(content_form, next_length))
                return form.copy(contents=contents, form_key=form_key)

            else:
                raise AssertionError(f"not a Form: {form!r}")

        return ak.operations.ak_from_buffers._impl(
            form=prepare(self, 1),
            length=1,
            container=container,
            buffer_key="{form_key}",
            backend=backend,
            byteorder=ak._util.native_byteorder,
            highlevel=False,
            behavior=behavior,
            attrs=None,
            simplify=False,
            enable_virtualarray_caching=True,
        )

    def _expected_from_buffers(
        self, getkey: Callable[[Form, str], str], recursive: bool
    ) -> Iterator[tuple[str, DType]]:
        raise NotImplementedError

    def expected_from_buffers(
        self, buffer_key="{form_key}-{attribute}", recursive=True
    ):
        """
        Args:
            buffer_key (str or callable): Python format string containing
                `"{form_key}"` and/or `"{attribute}"` or a function that takes these
                as keyword arguments and returns a string to use as a key for a buffer
                in the `container`.
            recursive (bool): If True, recurse into subforms; otherwise, yield
                only the (buffer_key, dtype) pairs for this form object.

        Yield (buffer_key, dtype) pairs describing the expected buffer keys,
        and their corresponding dtypes, that a call to #ak.from_buffers would
        be expected to find from the `container` object.
        """
        getkey = regularize_buffer_key(buffer_key)
        return dict(self._expected_from_buffers(getkey, recursive))

    def is_equal_to(
        self, other: Any, *, all_parameters: bool = False, form_key: bool = False
    ) -> bool:
        return self._is_equal_to(other, all_parameters, form_key)

    __eq__ = is_equal_to

    def _is_equal_to(self, other: Any, all_parameters: bool, form_key: bool) -> bool:
        raise NotImplementedError

    def _is_equal_to_generic(
        self, other: Any, all_parameters: bool, form_key: bool
    ) -> bool:
        compare_parameters = (
            parameters_are_equal if all_parameters else type_parameters_equal
        )

        return (
            isinstance(other, type(self))
            and not (form_key and self._form_key != other._form_key)
            and compare_parameters(self._parameters, other._parameters)
        )


def form_with_unique_keys(form: Form, key: tuple) -> Form:
    def impl(form: Form, key: tuple) -> None:
        # Set form key as str
        # Can be interpreted as tuple again with `ast.literal_eval(form_key)`
        form.form_key = repr(key)

        # If the form is a record we need to loop over all fields in the
        # record and set form that include the field name; this will keep
        # recursing as well.
        if form.is_record:
            for field in form.fields:
                impl(form.content(field), (*key, field))  # type: ignore[attr-defined]

        elif form.is_union:
            for i, entry in enumerate(form.contents):  # type: ignore[attr-defined]
                impl(entry, (*key, i))

        # NumPy like array is easy
        elif form.is_numpy or form.is_unknown:
            pass

        # Anything else grab the content and keep recursing
        else:
            # use `None` for anything else, could be more informative at some point, but there's no need for it right now
            impl(form.content, (*key, None))  # type: ignore[attr-defined]

    # Perform a "deep" copy without preserving references
    form = ak.forms.from_dict(form.to_dict())
    impl(form, key)
    return form



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext
from awkward._nplikes.numpy_like import NumpyMetadata

__all__ = ("real",)

np = NumpyMetadata.instance()


@ak._connect.numpy.implements("real")
@high_level_function()
def real(array, highlevel=True, behavior=None, attrs=None):
    """
    Args:
        array : array_like
            Input array.
        highlevel (bool, default is True): If True, return an #ak.Array;
            otherwise, return a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Returns the real components of the given array elements.
    If the arrays have complex elements, the returned arrays are floats.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, highlevel, behavior, attrs)


def _impl(array, highlevel, behavior, attrs):
    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array, allow_record=False, primitive_policy="error")

    out = ak._do.recursively_apply(layout, _action_real)
    return ctx.wrap(out, highlevel=highlevel)


def _action_real(layout, backend, **kwargs):
    if isinstance(layout, ak.contents.NumpyArray):
        return ak.contents.NumpyArray(backend.nplike.real(layout.data))
    else:
        return None



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import re

import awkward as ak
from awkward._dispatch import high_level_function

__all__ = ("from_tensorflow",)


@high_level_function()
def from_tensorflow(array):
    """
    Args:
        array: (TensorFlow Tensor):
            Tensor to convert into an Awkward Array.
    Converts a TensorFlow Tensor into an Awkward Array.
    If `array` contains any other data types the function raises an error.
    """

    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array)


def _impl(array):
    try:
        import tensorflow as tf
    except ImportError as err:
        raise ImportError(
            """to use ak.from_tensorflow, you must install the 'tensorflow' package with:

        pip install tensorflow
or
        conda install tensorflow"""
        ) from err

    # check if array is a Tensor
    if not isinstance(array, tf.Tensor):
        raise TypeError(
            """only a TensorFlow Tensor can be converted to Awkward Array"""
        )

    # keep the resulting array on the same device as input tensor
    device = array.backing_device
    matched_device = re.match(".*:(CPU|GPU):[0-9]+", device)

    if matched_device is None:
        raise NotImplementedError(
            f"TensorFlow device has an unexpected format: {device!r}"
        )
    elif matched_device.groups()[0] == "GPU":
        from awkward._nplikes.cupy import Cupy

        cp = Cupy.instance()
        # zero-copy data exchange through DLPack
        cp_array = cp.from_dlpack(tf.experimental.dlpack.to_dlpack(array))
        ak_array = ak.from_cupy(cp_array)

    elif matched_device.groups()[0] == "CPU":
        # this makes a copy unfortunately, since numpy is mutable and TensorFlow tensor is not
        np_array = array.numpy()
        ak_array = ak.from_numpy(np_array)

    return ak_array
